import numpy as np
import pandas as pd


class LinearRegressor():
    """
    Ordinary Least Squares Linear Regressor

    LinearRegressor will fit a linear model with coefficients w = (w1, w2, ..., wn)
    to minimize Residual Sum of Squares (RSS) between the observed targets values
    in the dataset, and the targets predicted by the linear approximation for the
    examples in the dataset.
    """
    def __init__(self):
        self.coef = None
        pass

    def fit(self, X, y, lambda_reg=0.1):
        X_np = np.array(X)
        y_np = np.array(y)

        # Check if dimensions of both matrices are correct
        if X_np.ndim != 2:
            raise ValueError("X should be a 2D array.")
        if y_np.ndim != 1:
            raise ValueError("y should be a 1D array.")

        # Check if the number of samples in X and y match
        if X_np.shape[0] != y_np.shape[0]:
            raise ValueError("The number of examples in X and y should be equal.")

        if X_np.shape[0] < X_np.shape[1]:
            raise ValueError("The number of examples in X should be greater than the number of features.")

        # Add a column of ones for the intercept term
        X_np = np.hstack((np.ones((X_np.shape[0], 1)), X_np))

        # Fit OLS with regularization (you can adjust the regularization parameter)
        self.coef = np.linalg.inv(X_np.T @ X_np + lambda_reg * np.eye(X_np.shape[1])) @ X_np.T @ y_np

        return self.coef

    def predict(self, X):
        """
        Predicts target values using the fitted linear model.
        
        Parameters
        ----------
        X : array-like matrix of shape (n_samples, n_features)
            Feature values that will be used to make predictions.
        
        Returns
        -------
        predictions : array-like matrix of shape (n_samples, n_targets)
            Predicted target values for the input feature values.
        """
        if self.coef is None:
            raise ValueError("Model not fitted. Call fit first.")

        X = np.array(X)

        # Check if dimensions of X are correct
        if X.ndim != 2:
            raise ValueError("X should be a 2D array.")

        # Check if the number of features in X equals to the number of coefficients
        if X.shape[1] != len(self.coef)-1:
            raise ValueError("The number of features in X should be equal to the number of coefficients.")
        
        # Check if non-numeric values exist in input
        if not np.issubdtype(X.dtype, np.number):
            raise ValueError("Input contains non-numeric values.")
            
        # Check if NaN values exist in input
        if np.isnan(X).any():
            raise ValueError("Input contains NaN values.")
        
        # check if infinite values exist in input
        if not np.isfinite(X).all():
            raise ValueError("Input contains infinite values.")

        X = np.hstack((np.ones((X.shape[0], 1)), X))

        pred = X @ self.coef
        return pred


    def score(self, X, y):
        """
        Calculates the coefficient of determination R^2 for the prediction.

        Parameters
        ----------
        X : array-like matrix, shape (n_samples, n_features)
            Feature dataset.

        y : array-like matrix, shape (n_samples, )
            True target values.

        Returns
        -------
        r2_score : float
            Coefficient of determination R^2.
        """
        # Ensure y is a numpy array
        y_true = np.array(y)

        # Predict the y values using the model
        y_pred = self.predict(X)

        # Calculate the mean of the true y values
        y_true_mean = np.mean(y_true)

        # Calculate the Total Sum of Squares (SST)
        SST = np.sum((y_true - y_true_mean) ** 2)

        # Calculate the Sum of Squared Errors (SSE)
        SSE = np.sum((y_true - y_pred) ** 2)

        # Calculate R^2
        r2 = 1 - (SSE / SST)

        return r2




from numpy import random
import numpy as np
import time


def cross_validate(model, X, y, cv=5, random_state=None):
    """
    Perform cross-validated Ordinary Least Squares (OLS) regression.

    Parameters
    ----------
    model : str
            Name of the model to run cross_validate with 
            (it will be OLS in this case)
    
    X : array-like matrix of shape (n_examples, n_features)
            Dataset that will be used as the feature values to train the model.
    
    y : array-like matrix of shape (n_examples, n_targets)
            Dataset that will be used as the target values to train the model.
    
    cv : (int, optional)
            Number of cross-validation folds. Default is 5.
    
    random_state : (int or None, optional) 
                    Seed for reproducibility. Default is None.


    Returns
    -------
    scores : A dict of arrays containing the score/time arrays for 
    each scorer is returned.
    """
    if not isinstance(model, object):
        raise TypeError("model is not a valid regression model instance")
    if not hasattr(model, "fit") or not hasattr(model, "score"):
        raise TypeError("model is not a valid regression model instance")
    if not isinstance(X, np.ndarray):
        raise ValueError("X is not a numpy array")
    if not isinstance(y, np.ndarray):
        raise ValueError("y is not a numpy array")
    if X.shape[0] != y.shape[0]:
        raise ValueError(f"The shape of X {X.shape} is not compatible with the shape of y {y.shape}")
    if type(cv) is not int or cv <= 0:
        raise ValueError("cross validation fold must be a positive integer")
    random.seed(random_state)
    data_combined = np.column_stack((X, y))
    np.random.shuffle(data_combined)
    batch_size = data_combined.shape[0] // cv
    batches = [
        data_combined[i: i + batch_size]
        for i in range(0, data_combined.shape[0], batch_size)
    ]
    train_score, test_score = [], []
    fit_time, score_time = [], []
    for i in range(len(batches)):
        train_batches = np.vstack(batches[:i] + batches[i+1:])
        t = time.time()
        model.fit(train_batches[:, :-1], train_batches[:, -1])
        fit_time.append(time.time() - t)
        t = time.time()
        test_score.append(model.score(batches[i][:, :-1], batches[i][:, -1]))
        score_time.append(time.time() - t)
        train_score.append(
            model.score(train_batches[:, :-1], train_batches[:, -1])
        )
    return {
        "train_score": train_score,
        "test_score": test_score,
        "fit_time": fit_time,
        "score_time": score_time,
    }


